/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.vauchannel.cxf;

import de.gematik.ti.vauchannel.protocol.TransportedData;
import de.gematik.ti.vauchannel.protocol.VAUProtocol;
import de.gematik.ti.vauchannel.protocol.VAUProtocolException;
import de.gematik.ti.vauchannel.protocol.VAUProtocolSession;
import org.apache.commons.codec.Charsets;
import org.apache.cxf.binding.soap.SoapFault;
import org.apache.cxf.interceptor.Fault;
import org.apache.cxf.io.CachedOutputStream;
import org.apache.cxf.message.Message;
import org.apache.cxf.phase.AbstractPhaseInterceptor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static de.gematik.ti.vauchannel.cxf.SOAPHelper.writeErrorStringToMessage;
import static de.gematik.ti.vauchannel.cxf.VAUMessageContext.*;


public class AESInterceptor extends AbstractPhaseInterceptor<Message> {

    private final Logger log = LoggerFactory.getLogger(this.getClass());
    @Autowired
    VAUProtocolProvider vauProtocolProvider;


    public AESInterceptor(String phase) {
        super(phase);
    }

    public void handleMessage(Message message) {

        String errorMessage = (String) message.getExchange().get("jsonerror");
        if (errorMessage == null) {
            try {
                handleVauProtocolAES(message);
            } catch (org.apache.cxf.transport.http.HTTPException he) {
                if (he.getResponseCode() == 403) {
                    log.error("got HTTP status code 403 -> access denied");
                }
                Fault f = new SoapFault("Access Denied", SoapFault.FAULT_CODE_CLIENT);
                throw f;
            } catch (Exception e) {
                log.error(e.getMessage(), e);
                handleExceptionInEncryptionLayer(message, e.getMessage());
            }
        } else {
            try {
                handleVauProtocolError(message, errorMessage);
            } catch (Exception e) {
                log.error(e.getMessage(), e);
            }
        }
    }


    public void handleFault(Message message) {
    }

    public void handleVauProtocolAES(Message message) throws Exception {

        VAUProtocol vauProtocol = vauProtocolProvider.getVAUProtocol(message);


        VAUProtocolSession session = vauProtocol.session();
        String contentType = (String) message.get("Content-Type");


        if (session.isClient()) {
            Object o = message.get("org.apache.cxf.message.inbound");
            if (o != null && (Boolean) o) {
                if (contentType != null && contentType.startsWith("application/json")) {
                    byte[] rawData = readContentFromMessage(message);
                    String serverError = new String(rawData);
                    log.info(serverError);
                    String errorString = vauProtocol.validateAndUnpackServerError(serverError);
                    log.info(errorString);
                    writeErrorStringToMessage(errorString, message);

                }
            }
        }

        // [A_16884] - VAU-Protokoll: Nachrichtentypen und HTTP-Content-Type

        if (contentType != null && contentType.startsWith("application/json")) {
            return;
        }

        boolean isOutbound = message == message.getExchange().getOutMessage()
                || message == message.getExchange().getOutFaultMessage();


        VAUMessageContext vauMessageContext = VAU_CHANNEL_NOT_ALLOWED;
        if (contentType != null && contentType.startsWith("application/json")) {
            vauMessageContext = VAU_CHANNEL_HANDSHAKE;
        } else if (isOutbound) {
            vauMessageContext = VAU_CHANNEL_TRANSPORT;
        } else if (contentType != null && contentType.startsWith("application/octet-stream")) {
            vauMessageContext = VAU_CHANNEL_TRANSPORT;
        }

        if (vauMessageContext == VAU_CHANNEL_NOT_ALLOWED) {
            String errorMessage = VAUProtocolException.ACCESS_DENIED;
            log.error(errorMessage);
            handleExceptionInEncryptionLayer(message, errorMessage);
        }

        if (isOutbound) {
            if (session.isClient() && vauMessageContext == VAU_CHANNEL_TRANSPORT && session.getSymKey() == null) {
                throw new VAUProtocolException(VAUProtocolException.ACCESS_DENIED);
            }
            if (!session.isClient() && session.getCounter() == 0) {
                session.setCounter(0);
            }
            OutputStream os = message.getContent(OutputStream.class);
            CachedStream cs = new CachedStream();
            message.setContent(OutputStream.class, cs);
            message.getInterceptorChain().doIntercept(message);

            try {
                cs.flush();

                String originalContentType = (String) message.get("Content-Type");
                Map<String, List> headers = (Map<String, List>) message.get(Message.PROTOCOL_HEADERS);
                if (headers == null) {
                    headers = new HashMap<String, List>();
                    message.put(Message.PROTOCOL_HEADERS, headers);
                }

                message.put("Content-Type", "application/octet-stream");  // [A_16884]

                CachedOutputStream csnew = (CachedOutputStream) message
                        .getContent(OutputStream.class);

                message.setContent(OutputStream.class, os);

                byte[] rawMessageData = csnew.getBytes();
                log.debug("rawMessageData: " + new String(rawMessageData, StandardCharsets.UTF_8));
                byte[] finalMessageData = null;

                if (!session.isClient() && isAccessDeniedException(rawMessageData)) {
                    String errorStr = vauProtocol.generateVAUServerErrorMessage(VAUProtocolException.ACCESS_DENIED);
                    finalMessageData = errorStr.getBytes(Charsets.UTF_8);
                    message.put("Content-Type", "application/json");  // [A_16884]
                    message.put(Message.RESPONSE_CODE, 403);
                    session.setActive(false);
                } else {
                    TransportedData transportedData = new TransportedData(rawMessageData, originalContentType);
                    finalMessageData = vauProtocol.encrypt(transportedData);
                }
                if (!session.isClient() && !session.isActive()) {
                    this.vauProtocolProvider.closeVAUProtocol(vauProtocol);
                }
                os.write(finalMessageData, 0, finalMessageData.length);
            } catch (IOException e) {
                if (session != null) {
                    log.error(e.getMessage(), e);
                }
            } finally {
                try {
                    cs.close();
                    os.flush();
                    os.close();
                } catch (org.apache.cxf.transport.http.HTTPException he) {
                    if (session.isClient() && he.getResponseCode() == 403) {
                        log.error("got HTTP status code 403 -> access denied");
                    }
                    throw he;
                } catch (Exception e) {
                    log.error(e.getMessage(), e);
                }
            }
        } else {
            byte[] rawRequest = readContentFromMessage(message);
            TransportedData transportedData = vauProtocol.decrypt(rawRequest);
            message.put("Content-Type", transportedData.contentType);
            InputStream myInputStream = new ByteArrayInputStream(transportedData.body);
            message.setContent(InputStream.class, myInputStream);
        }
    }

    private boolean isAccessDeniedException(byte[] rawMessageData) {
        try {
            String xml = new String(rawMessageData, Charsets.UTF_8);
            // simplified
            return xml.indexOf(VAUProtocolException.ACCESS_DENIED) != -1;
        } catch (Exception e) {
        }
        return false;
    }


    public byte[] readContentFromMessage(Message message) {
        InputStream is = message.getContent(InputStream.class);

        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        try {
            org.apache.cxf.helpers.IOUtils.copy(is, bout);
        } catch (IOException ex) {
            log.error(ex.getMessage(), ex);
        }
        byte[] rawRequest = bout.toByteArray();
        return rawRequest;
    }

    private void handleExceptionInEncryptionLayer(Message message, String errorMessage) {
        VAUProtocol vauProtocol = vauProtocolProvider.getVAUProtocol(message);
        if (!vauProtocol.session().isClient()) {
            message.put(Message.RESPONSE_CODE, Response.Status.BAD_REQUEST.getStatusCode());
            message.getExchange().put("jsonerror", errorMessage);
            log.error("Exception in Encryption Layer");
            throw new RuntimeException(errorMessage);
        } else {
            log.error(errorMessage);
        }
    }

    private void handleVauProtocolError(Message message, String jsonerror) {
        message.put(HttpHeaders.CONTENT_TYPE, MediaType.APPLICATION_JSON);

        VAUProtocol vauProtocol = vauProtocolProvider.getVAUProtocol(message);
        VAUProtocolSession session = vauProtocol.session();
        if (!session.isClient() && !session.isActive()) {
            this.vauProtocolProvider.closeVAUProtocol(vauProtocol);
        }

        String errorStr = vauProtocol.generateVAUServerErrorMessage(jsonerror);

        OutputStream os = message.getContent(OutputStream.class);
        CachedStream cs = new CachedStream();
        message.setContent(OutputStream.class, cs);

        message.getInterceptorChain().doIntercept(message);

        try {
            cs.flush();
            CachedOutputStream csnew = (CachedOutputStream) message
                    .getContent(OutputStream.class);

            message.setContent(OutputStream.class, os);
            csnew.getBytes();
            byte[] data_ = errorStr.getBytes(StandardCharsets.UTF_8);
            os.write(data_, 0, data_.length);

            cs.close();
            os.flush();
            os.close();

        } catch (IOException e) {
            log.error(e.getMessage(), e);
        }
    }


    private class CachedStream extends CachedOutputStream {

        public CachedStream() {
            super();
        }

        protected void doFlush() throws IOException {
            currentStream.flush();
        }

        protected void doClose() throws IOException {
        }

        protected void onWrite() throws IOException {
        }
    }

}
